/*
 * Copyright 2020 sfner.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package br.ufes.inf.decisionelement.wheelreinventions.expressionparsing;

import br.ufes.inf.decisionelement.wheelreinventions.expressionparsing.tokenization.MathExpressionException;
import br.ufes.inf.decisionelement.wheelreinventions.expressionparsing.tokenization.AdditionOperatorExpressionParser;
import br.ufes.inf.decisionelement.wheelreinventions.expressionparsing.tokenization.LiteralExpressionParser;
import br.ufes.inf.decisionelement.wheelreinventions.expressionparsing.tokenization.ContextualizationExpressionParser;
import br.ufes.inf.decisionelement.wheelreinventions.expressionparsing.tokenization.DivisionOperatorExpressionParser;
import br.ufes.inf.decisionelement.wheelreinventions.expressionparsing.tokenization.MultiplicationOperatorExpressionParser;
import br.ufes.inf.decisionelement.wheelreinventions.expressionparsing.tokenization.SubtractionOperatorExpressionParser;
import br.ufes.inf.decisionelement.wheelreinventions.expressionparsing.tokenization.VariableExpressionParser;
import br.ufes.inf.decisionelement.wheelreinventions.expressionparsing.tokenization.ParenthesisExpressionParser;
import br.ufes.inf.decisionelement.wheelreinventions.CharArrayUtils;
import br.ufes.inf.decisionelement.wheelreinventions.positionalcollections.Tuple;
import java.util.ArrayList;
import java.util.List;

/**
 * No docstring set.
 * @param <T>
 */
public abstract class MathExpressionParser<T> {
    public final T token;
    public MathExpressionParser() {
        token = null;
    }
    public MathExpressionParser(T t) {
        token = t;
    }
    private static final MathExpressionParser<?>[] EXPRESSION_PARSERS = new MathExpressionParser<?>[]{
        new ParenthesisExpressionParser(),
        new VariableExpressionParser(),
        new LiteralExpressionParser(),
        new AdditionOperatorExpressionParser(),
        new SubtractionOperatorExpressionParser(),
        new MultiplicationOperatorExpressionParser(),
        new DivisionOperatorExpressionParser(),
        new ContextualizationExpressionParser(),
    };
    protected final boolean charSequenceSpans = getCharsSequenceSpans();
    protected final char[] recognizedChars = getRecognizedChars();
    protected final int consumedPrecedingTokens = getConsumedPrecedingTokens();
    protected final int consumedSuccedingTokens = getConsumedSuccedingTokens();
    protected final boolean isConsumable = getIsConsumable();
    protected abstract boolean getCharsSequenceSpans();
    protected abstract char[] getRecognizedChars();
    protected abstract int getConsumedPrecedingTokens();
    protected abstract int getConsumedSuccedingTokens();
    protected abstract boolean getIsConsumable();
    protected Tuple<String, String> attemptDefaultTokenization(String expression) {
        expression = expression.stripLeading();
        if (expression.length() > 0 && CharArrayUtils.contains(recognizedChars, expression.charAt(0))) {
            int i = 1;
            if (!charSequenceSpans) {
                i--;
            } else {
                for (; i < expression.length(); i++) {
                    if (!CharArrayUtils.contains(recognizedChars, expression.charAt(i))) {
                        i--;
                        break;
                    }
                }
                if (i == expression.length()) {
                    i--;
                }
            }
            return new Tuple<>(expression.substring(0, i + 1).strip(), expression.substring(i + 1).stripLeading());
        }
        return null;
    }
    protected abstract Tuple<MathExpressionParser<?>, String>
        tryTokenization(String expression);
    protected abstract Tuple<List<MathExpressionParser<?>>, Integer>
        parseTokens(List<MathExpressionParser<?>> tokens, int i);
    public static Tuple<List<MathExpressionParser<?>>, String>
        parsePartial(String expression) throws MathExpressionException {
        final List<MathExpressionParser<?>> tokens = new ArrayList<>();
        String remainingExpression = expression.stripLeading();
        boolean tokenFound = true;
        while (tokenFound) {
            tokenFound = false;
            for (final var expressionParser: EXPRESSION_PARSERS) {
                final Tuple<MathExpressionParser<?>, String> resp = expressionParser.
                        tryTokenization(remainingExpression);
                if (resp != null && resp.first != null) {
                    tokenFound = true;
                    remainingExpression = resp.second.stripLeading();
                    tokens.add(resp.first);
                    break;
                }
            }
        }
        if (remainingExpression.stripLeading().length() > 0) {
            throw new MathExpressionException("Cannot tokenize '" + remainingExpression + "'");
        }
        return new Tuple<>(tokens, remainingExpression.stripLeading());
    }
    public static List<MathExpressionParser<?>> parse(String expression) throws MathExpressionException {
        return parsePartial(expression).first;
    }
    @Override
    public String toString() {
        return "" + token;
    }
}
